''' SRCDS Types Python module '''

import vecmath


def tupleconvert(method):
    '''
    Decorator for methods making sure the parameter is a tuple.
    '''
    def wrapper_caller(self, stuff):
        if not isinstance(stuff, tuple):
            stuff = stuff,
        method(self, stuff)
    return wrapper_caller


class Vector(vecmath.Vector):
    def getvalue(self):
        '''Return the value used for setting keyvalues.'''
        return self.getstr(' ')


class Color(object):
    '''Class for storing entity colors.'''
    def __init__(self, red=0, green=0, blue=0, alpha=255, color=0):
        if isinstance(red, (tuple, list)):
            return self.__init__(*red)
        self.red = red & 0xff
        self.green = green & 0xff
        self.blue = blue & 0xff
        self.alpha = alpha & 0xff
        if color:
            self.red, self.green, self.blue, self.alpha = color_tuple(color)
    
    def __repr__(self):
        return 'Color(%03d, %03d, %03d, %03d)' % (
            self.red,
            self.green,
            self.blue,
            self.alpha,
        )

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        return int(self) == int(other)
    
    def __str__(self):
        '''str(x), return the color in usable format for the engine.'''
        return '%d %d %d %d'%(self.red, self.green, self.blue, self.alpha)

    def __int__(self):
        '''int(x), return the color as a single integer.'''
        return (
            (self.blue) +
            (self.green << 8) + 
            (self.red << 16) + 
            (self.alpha << 24)
        )
	
    def __call__(self, *args, **kw):
        self.__init__(*args, **kw)

    def __tuple__(self):
        '''tuple(x), return the color information as a tuple.'''
        return self.red, self.green, self.blue, self.alpha

    def __list__(self):
        '''list(x), return the color information as a list.'''
        return [self.red, self.green, self.blue, self.alpha]

    def setvalue(self, value):
        '''Set a new value.'''
        try:
            clrRender = int(value)
            self.red, self.green, self.blue, self.alpha = color_tuple(clrRender)
        except ValueError:
            self.red, self.green, self.blue, self.alpha = (int(clr)&0xff for clr in value.split(' '))
        except TypeError:
            if len(value) == 3:
                self.red, self.green, self.blue = value
            elif len(value) == 4:
                self.red, self.green, self.blue, self.alpha = value
            else:
                raise TypeError, 'Invalid object %s for Color'%repr(value)

    def getvalue(self):
        '''Return the value used for setting keyvalues.'''
        return str(self)


def color_tuple(clrRender):
    '''
    Convert the single integer into a tuple of (red, gree, blue, alpha).
    '''
    blue = clrRender & 0xff
    green = (clrRender >> 8) & 0xff
    red = (clrRender >> 16) & 0xff
    alpha = (clrRender >> 24) & 0xff

    return red, green, blue, alpha


class RangedValue(object):
    '''
    A wrapper for any comparable data type that checks the value to be between
    given range of values.
    '''
    def __init__(self, value, minimum, maximum):
        self._value = None
        self._min = minimum
        self._max = maximum
        self.setvalue(value)
    
    def __repr__(self):
        return '%s(%r, %r, %r)' % (
            self.__class__.__name__,
            self._value,
            self._min,
            self._max,
        )
    
    def setvalue(self, value):
        '''Set a new value or raise ValueError if out of range.'''
        typedvalue = type(self._min)(value)
        if self._min <= typedvalue <= self._max:
            self._value = typedvalue
        else:
            raise ValueError, 'Value %s is out of range [%s - %s]'%(
                repr(typedvalue), repr(self._min), repr(self._max)
                )
        return self._value
    
    def getvalue(self):
        '''Get the current value.'''
        return self._value

    
class SymbolicValue(object):
    '''
    An object that allows using direct values and symbolic keyword values.
    '''
    def __init__(self, initial, keywords):
        self._value = None
        self._keywords = keywords
        self._values = set(keywords.values())
        if self._values:
            self._type = type(iter(self._values).next())
        else:
            self._type = int
        self.setvalue(initial)
    
    def __repr__(self):
        return '%s(%r, %r)' % (
            self.__class__.__name__,
            self._value,
            self._keywords,
        )
            
    def extend(self, instance):
        self._keywords.update(instance._keywords)
        self._values = set(self._keywords.values())
        self._value = instance._value
        
    def setvalue(self, value):
        '''Set a new value. '''
        try:
            testvalue = self._type(value)
        except (ValueError, TypeError), e:
            testvalue = value
        if testvalue in self._values:
            self._value = testvalue
        else:
            self._value = self._keywords[value]
        return self._value

    def getvalue(self):
        '''Get current value. '''
        return self._value


class SymbolicBoolean(SymbolicValue):
    '''SymbolicValue with predefined boolean values.'''
    def __init__(self, initial=0):
        super(SymbolicBoolean, self).__init__(initial, {
            'No': 0,
            'Yes': 1,
        })

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            ('No', 'Yes')[self._value]
        )
    
    def setvalue(self, value):
        if not isinstance(value, basestring):
            value = value and 'Yes' or 'No'
        super(SymbolicBoolean, self).setvalue(value)


class Flags(object):
    '''An object working as binary flags supporting also symbolic names.'''
    def __init__(self, flags, *initial):
        self._value = 0
        self._flags = flags
        self.set_flags(*initial)

    def extend(self, flags_instance):
        self._flags.update(flags_instance._flags)
        self._value |= flags_instance._value
        
    def __int__(self):
        return self._flags

    def __str__(self):
        return str(self._flags)

    def __repr__(self):
        symbolic_flags = ['']
        for flag_name, flag_value in self._flags.iteritems():
            if flag_value & self._value:
                symbolic_flags.append(flag_name)
        return '%s(%r%s)' % (
            self.__class__.__name__,
            self._flags,
            ', '.join(symbolic_flags),
        )

    @tupleconvert
    def __ior__(self, flags):
        self.edit_flags(lambda a,b:a|b, flags)
    
    @tupleconvert
    def __iadd__(self, flags):
        for flag in flags:
            if isinstance(flag, int):
                self._value += int(flag)
            else:
                self._value |= self._flags[flag]

    @tupleconvert
    def __ixor__(self, flags):
        self.edit_flags(lambda a,b: a^b, flags)

    @tupleconvert
    def __iand__(self, flags):
        self.edit_flags(lambda a,b: a&b, flags)

    @tupleconvert
    def __isub__(self, flags):
        for flag in flags:
            if isinstance(flag, int):
                self._value -= int(flag)
            else:
                self._value &= ~self._flags[flag]

    def setvalue(self, value):
        self._value = int(value)

    def edit_flags(self, op, flags):
        '''Edits flags using specified operator.'''
        for flag in flags:
            if isinstance(flag, int):
                nflag = int(flag) # instance of int is not necessarily int
            else:
                nflag = self._flags[flag] # look up by name
            self._value = op(self._value, nflag)

    def set_flags(self, *flags):
        '''Set flags on.'''
        self.edit_flags(lambda a,b: a|b, flags)

    def toggle_flags(self, *flags):
        '''Toggle flags.'''
        self.edit_flags(lambda a,b: a^b, flags)

    def clear_flags(self, *flags):
        '''Set flags off.'''
        self.edit_flags(lambda a,b: a&~b, flags)


class KeyvalueInput(object):
    '''
    A type that has an Input of the exact same name to change value.
    '''
    def __init__(self, inputcommand, value):
        self._inputcommand = inputcommand
        self._value = value

    def setvalue(self, value):
        if hasattr(self._value, 'setvalue'):
            self._value.setvalue(value)
        elif type(value) is not type(self._value):
            self._value = type(self._value)(value)
        else:
            self._value = value

    def __call__(self, entity, value, userid=0):
        self.setvalue(value)
        entity.fire('%s "%s"'%(self._inputcommand, self._value), userid)

